
package org.mule.modules.mvel.expressions;

import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import org.mule.api.MuleContext;
import org.mule.api.MuleException;
import org.mule.api.context.MuleContextAware;
import org.mule.api.expression.ExpressionEnricher;
import org.mule.api.lifecycle.Disposable;
import org.mule.api.lifecycle.Initialisable;
import org.mule.api.lifecycle.InitialisationException;
import org.mule.api.lifecycle.Startable;
import org.mule.api.lifecycle.Stoppable;
import org.mule.api.transformer.DataType;
import org.mule.api.transformer.Transformer;
import org.mule.api.transformer.TransformerException;
import org.mule.api.transport.PropertyScope;
import org.mule.expression.ExpressionUtils;
import org.mule.modules.mvel.adapters.MVELExpressionLanguageInjectionAdapter;
import org.mule.transformer.types.DataTypeFactory;

public class MvelExpressionEnricher
    implements MuleContextAware, ExpressionEnricher, Disposable, Initialisable, Startable, Stoppable
{

    private MVELExpressionLanguageInjectionAdapter module = null;

    public MvelExpressionEnricher() {
        module = new MVELExpressionLanguageInjectionAdapter();
    }

    public void setMuleContext(MuleContext muleContext) {
        if (module instanceof MuleContextAware) {
            ((MuleContextAware) module).setMuleContext(muleContext);
        }
    }

    public void start()
        throws MuleException
    {
        if (module instanceof Startable) {
            ((Startable) module).start();
        }
    }

    public void stop()
        throws MuleException
    {
        if (module instanceof Stoppable) {
            ((Stoppable) module).stop();
        }
    }

    public void initialise()
        throws InitialisationException
    {
        if (module instanceof Initialisable) {
            ((Initialisable) module).initialise();
        }
    }

    public void dispose() {
        if (module instanceof Disposable) {
            ((Disposable) module).dispose();
        }
    }

    public String getName() {
        return "mvel";
    }

    public void setName(String name) {
        throw new UnsupportedOperationException();
    }

    /**
     * Get all superclasses and interfaces recursively.
     * 
     * @param classes List of classes to which to add all found super classes and interfaces.
     * @param clazz   The class to start the search with.
     */
    private void computeClassHierarchy(Class clazz, List classes) {
        for (Class current = clazz; (current!= null); current = current.getSuperclass()) {
            if (classes.contains(current)) {
                return ;
            }
            classes.add(current);
            for (Class currentInterface: current.getInterfaces()) {
                computeClassHierarchy(currentInterface, classes);
            }
        }
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link List }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isListClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(List.class);
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link Map }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isMapClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(Map.class);
    }

    private boolean isList(Type type) {
        if ((type instanceof Class)&&isListClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isList(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isList(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isMap(Type type) {
        if ((type instanceof Class)&&isMapClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isMap(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isMap(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isAssignableFrom(Type expectedType, Class clazz) {
        if (expectedType instanceof Class) {
            if (((Class) expectedType).isPrimitive()) {
                if (((Class) expectedType).getName().equals("boolean")&&(clazz == Boolean.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("byte")&&(clazz == Byte.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("short")&&(clazz == Short.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("char")&&(clazz == Character.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("int")&&(clazz == Integer.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("float")&&(clazz == Float.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("long")&&(clazz == Long.class)) {
                    return true;
                }
                if (((Class) expectedType).getName().equals("double")&&(clazz == Double.class)) {
                    return true;
                }
                return false;
            } else {
                return ((Class) expectedType).isAssignableFrom(clazz);
            }
        }
        if (expectedType instanceof ParameterizedType) {
            return isAssignableFrom(((ParameterizedType) expectedType).getRawType(), clazz);
        }
        if (expectedType instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) expectedType).getUpperBounds();
            if (upperBounds.length!= 0) {
                return isAssignableFrom(upperBounds[ 0 ], clazz);
            }
        }
        return false;
    }

    private Object transform(org.mule.api.MuleMessage muleMessage, Type expectedType, Object source)
        throws TransformerException
    {
        if (source == null) {
            return source;
        }
        Object target = null;
        if (isList(source.getClass())) {
            if (isList(expectedType)) {
                List newList = new ArrayList();
                Type valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                ListIterator iterator = ((List) source).listIterator();
                while (iterator.hasNext()) {
                    Object subTarget = iterator.next();
                    newList.add(transform(muleMessage, valueType, subTarget));
                }
                target = newList;
            } else {
                target = source;
            }
        } else {
            if (isMap(source.getClass())) {
                if (isMap(expectedType)) {
                    Type keyType = Object.class;
                    Type valueType = Object.class;
                    if (expectedType instanceof ParameterizedType) {
                        keyType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                        valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 1 ];
                    }
                    Map map = ((Map) source);
                    Map newMap = new HashMap();
                    for (Object entryObj: map.entrySet()) {
                        {
                            Map.Entry entry = ((Map.Entry) entryObj);
                            Object newKey = transform(muleMessage, keyType, entry.getKey());
                            Object newValue = transform(muleMessage, valueType, entry.getValue());
                            newMap.put(newKey, newValue);
                        }
                    }
                    target = newMap;
                } else {
                    target = source;
                }
            } else {
                target = source;
            }
        }
        if ((target!= null)&&(!isAssignableFrom(expectedType, target.getClass()))) {
            DataType sourceDataType = DataTypeFactory.create(target.getClass());
            if (expectedType instanceof ParameterizedType) {
                expectedType = ((ParameterizedType) expectedType).getRawType();
            }
            DataType targetDataType = DataTypeFactory.create(((Class) expectedType));
            Transformer t = muleMessage.getMuleContext().getRegistry().lookupTransformer(sourceDataType, targetDataType);
            return t.transform(target);
        } else {
            return target;
        }
    }

    public void enrich(String expression, org.mule.api.MuleMessage message, Object object) {
        try {
            Class[] parameterClasses = new Class[] {String.class, Object.class, org.mule.api.MuleMessage.class, Map.class, Map.class, Map.class, Map.class };
            Method evaluateMethod = module.getClass().getMethod("enrich", parameterClasses);
            Type expressionType = evaluateMethod.getGenericParameterTypes()[ 0 ];
            Type valueType = evaluateMethod.getGenericParameterTypes()[ 1 ];
            Type muleMessageType = evaluateMethod.getGenericParameterTypes()[ 2 ];
            Type inboundHeadersType = evaluateMethod.getGenericParameterTypes()[ 3 ];
            Type outboundHeadersType = evaluateMethod.getGenericParameterTypes()[ 4 ];
            Type invocationHeadersType = evaluateMethod.getGenericParameterTypes()[ 5 ];
            Type sessionHeadersType = evaluateMethod.getGenericParameterTypes()[ 6 ];
            MvelExpressionEnricher.TrackMap<String, Object> inboundHeaders = null;
            inboundHeaders = new MvelExpressionEnricher.TrackMap<String, Object>(((Map<String, Object> ) transform(message, inboundHeadersType, ExpressionUtils.getPropertyWithScope("INBOUND:*", message, Map.class))));
            MvelExpressionEnricher.TrackMap<String, Object> outboundHeaders = null;
            outboundHeaders = new MvelExpressionEnricher.TrackMap<String, Object>(((Map<String, Object> ) transform(message, outboundHeadersType, ExpressionUtils.getPropertyWithScope("OUTBOUND:*", message, Map.class))));
            MvelExpressionEnricher.TrackMap<String, Object> invocationHeaders = null;
            invocationHeaders = new MvelExpressionEnricher.TrackMap<String, Object>(((Map<String, Object> ) transform(message, invocationHeadersType, ExpressionUtils.getPropertyWithScope("INVOCATION:*", message, Map.class))));
            MvelExpressionEnricher.TrackMap<String, Object> sessionHeaders = null;
            sessionHeaders = new MvelExpressionEnricher.TrackMap<String, Object>(((Map<String, Object> ) transform(message, sessionHeadersType, ExpressionUtils.getPropertyWithScope("SESSION:*", message, Map.class))));
            Object newPayload = module.enrich(expression, object, message, inboundHeaders, outboundHeaders, invocationHeaders, sessionHeaders);
            message.setPayload(newPayload);
            for (String key: inboundHeaders.changedKeySet()) {
                message.setProperty(key, inboundHeaders.get(key), PropertyScope.INBOUND);
            }
            for (String key: outboundHeaders.changedKeySet()) {
                message.setProperty(key, outboundHeaders.get(key), PropertyScope.OUTBOUND);
            }
            for (String key: sessionHeaders.changedKeySet()) {
                message.setProperty(key, sessionHeaders.get(key), PropertyScope.SESSION);
            }
            for (String key: invocationHeaders.changedKeySet()) {
                message.setProperty(key, invocationHeaders.get(key), PropertyScope.INVOCATION);
            }
        } catch (NoSuchMethodException e) {
            throw new RuntimeException(e);
        } catch (TransformerException e) {
            throw new RuntimeException(e);
        }
    }

    private class TrackMap<K, V >
        implements Map<K, V>
    {

        private Map<K, V> trackedMap;
        private Set<K> changedKeys;

        public TrackMap(Map<K, V> trackedMap) {
            this.trackedMap = trackedMap;
            this.changedKeys = new HashSet<K>();
        }

        @Override
        public int size() {
            return trackedMap.size();
        }

        @Override
        public boolean isEmpty() {
            return trackedMap.isEmpty();
        }

        @Override
        public boolean containsKey(Object o) {
            return trackedMap.containsKey(o);
        }

        @Override
        public boolean containsValue(Object o) {
            return trackedMap.containsValue(o);
        }

        @Override
        public V get(Object o) {
            return trackedMap.get(o);
        }

        @Override
        public V put(K k, V v) {
            changedKeys.add(k);
            return trackedMap.put(k, v);
        }

        @Override
        public V remove(Object o) {
            return trackedMap.remove(o);
        }

        @Override
        public void putAll(Map<? extends K, ? extends V> map) {
            trackedMap.putAll(map);
        }

        @Override
        public void clear() {
            trackedMap.clear();
        }

        @Override
        public Set<K> keySet() {
            return trackedMap.keySet();
        }

        public Set<K> changedKeySet() {
            return this.changedKeys;
        }

        @Override
        public Collection<V> values() {
            return trackedMap.values();
        }

        @Override
        public Set<Map.Entry<K, V>> entrySet() {
            return trackedMap.entrySet();
        }

    }

}
